#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import datetime
import logging
import numpy as np
import pandas as pd
import copy
from enum import Enum
from wiliot_core.packet_data.config_files.packet_map import parse_packet, gw_result_dict, construct_packet_from_fields
from wiliot_core.utils.utils import valid_packet_start
from wiliot_core.packet_data.config_files.packet_flow_parameters import get_flow_param


FULL_PACKET_PREFIX = 'full_packet'
LEGACY_PACKET_PREFIX = 'process_packet'


packet_tag_length = 74
MAX_STAT_PARAM_VAL = 65535  # 2 bytes
MAX_GW_CLOCK = 32767  # 7 bits


class InlayTypes(Enum):
    TEO_086 = '086'
    TIKI_096 = '096'
    TIKI_099 = '099'
    BATTERY_107 = '107'
    TIKI_117 = '117'
    TIKI_118 = '118'
    TIKI_121 = '121'
    TIKI_122 = '122'
    TIKI_136 = '136'
    TIKI_H_160 = '160'
    TIKI_160 = '160'
    TIKI_168 = '168'
    TIKI_169 = '169'


class Packet(object):
    """
    Wiliot Packet Object

        :param raw_packet: the raw packet to create a Packet object
        :type raw_packet: str or dict

        :return:
    """

    def __init__(self, raw_packet, time_from_start=None, custom_data=None,
                 inlay_type=None, logger_name=None, ignore_crc=False):
        """
        :param raw_packet:
        :type raw_packet: str or dict
        :param logger_name: the log file we want to write into it.
        :type time_from_start: str
        :param time_from_start: the time when the packet was received according to the gw timer
        :type time_from_start: float
        :param custom_data: the packet custom data as dictionary
        :type custom_data: dict
        :param inlay_type: the antenna/ inlay type (e.g. tiki, teo)
        :type inlay_type: str
        :type ignore_crc: bool
        :param ignore_crc: if True, update for all packets crc_valid = True,
                           relevant for data collected by old GW FW version
        """

        self.is_valid_packet = False
        if logger_name is not None:
            self.logger = logging.getLogger(logger_name)
        else:
            self.logger = None

        # init packet input:
        packet_in = None
        time_in = None
        self.gw_process = False
        if type(raw_packet) is dict and raw_packet['is_valid_tag_packet']:
            packet_in = raw_packet['packet']
            time_in = raw_packet['time_from_start']
            self.gw_process = True
        elif type(raw_packet) is str:
            packet_in = raw_packet
            time_in = time_from_start

        # check packet type:
        if FULL_PACKET_PREFIX in packet_in:
            self.is_full_packet = True
        elif LEGACY_PACKET_PREFIX in packet_in:
            self.is_full_packet = False
        else:
            self.is_full_packet = None

        # parse packet
        try:
            packet_in_strip = valid_packet_start(packet_in) or packet_in
            self.packet_data, self.gw_data = parse_packet(packet_in_strip, self.is_full_packet, ignore_crc)

            # assign results
            self.is_valid_packet = self.gw_data['is_valid_tag_packet'].take(0)
            self.gw_data['time_from_start'] = np.array(time_in)
            self.packet_data['is_full_packet'] = self.is_full_packet
        except Exception as e:
            self.printing_and_logging(f'Could not parse the following packet {packet_in} due to {e}')
            self.is_valid_packet = False
            self.gw_data = {'is_valid_tag_packet': np.array(False)}

        # add more info to packet struct
        self.decoded_data = {}
        if custom_data is not None:
            self.custom_data = custom_data
        else:
            self.custom_data = {}
        if 'timestamp' not in self.custom_data.keys():
            self.custom_data['timestamp'] = datetime.datetime.now().timestamp() * 1000  # ms
        self.inlay_type = inlay_type

        if self.is_valid_packet:
            enc_payload = self.packet_data['enc_payload']
            if len(enc_payload) == 0:
                self.is_valid_packet = False
                self.printing_and_logging(f'the following packet marked as invalid due to '
                                          f'empty tag data payload: {self.packet_data["raw_packet"]}')
                return
            if self.packet_data['enc_payload'] == \
                    self.packet_data['enc_payload'][0] * len(self.packet_data['enc_payload']):
                self.is_valid_packet = False
                self.printing_and_logging(f'the following packet marked as invalid due to identical characters in '
                                          f'the tag data payload: {self.packet_data["raw_packet"]}')
                return
            # clear adi from packet and add it to custom data
            if self.packet_data['ble_type'] != 'LEGACY':
                self.custom_data['adi_per_packet'] = [self.packet_data['adi']]
                self.packet_data['raw_packet'] = construct_packet_from_fields(packet_data=self.packet_data,
                                                                              fields_to_zero=['adi'])

    def __len__(self):
        """
        gets number of sprinkler occurrences in packet
        """
        return self.gw_data['rssi'].size

    def __eq__(self, packet):
        """
        packet comparison
        """
        if self.is_same_sprinkler(packet):
            if len(packet) == self.__len__():
                return all([np.take(packet.gw_data['gw_packet'], i) == np.take(self.gw_data['gw_packet'], i)
                            for i in range(len(packet))])
        return False

    def __str__(self):
        """
        packet print method
        """
        return str(
            'packet_data={packet_data}, gw_data={gw_data}'.format(packet_data=self.packet_data, gw_data=self.gw_data))

    def printing_and_logging(self, message):
        if self.logger:
            self.logger.info(message)
        else:
            print(message)

    def is_packet_from_bridge(self):
        return np.take(self.gw_data['is_packet_from_bridge'], 0)

    def get_payload(self):
        if self.is_valid_packet:
            return self.packet_data['data_uid'] + self.packet_data['raw_group_id'] + self.packet_data['nonce'] + \
                   self.packet_data['enc_uid'] + self.packet_data['mic'] + self.packet_data['enc_payload']
        else:
            return ''

    def get_group_id(self):
        return self.packet_data['group_id']

    def set_inlay_type(self, inlay_type):
        inlay_type = inlay_type if isinstance(inlay_type, str) else inlay_type.value
        self.inlay_type = inlay_type

    def is_in(self, packet):
        """
        is packet contains another packet

        :param packet: the other packet to verify
        :type packet: Packet

        :return: bool
        """
        if self.is_same_sprinkler(packet):
            if np.take(packet.gw_data['gw_packet'], 0) in self.gw_data['gw_packet']:
                return True
        return False

    def get_packet(self):
        """
        gets raw packet string
        """
        return str(self.packet_data['raw_packet'])

    def split_packet(self, index):
        """
        split packet by index
        """
        packet_a = self.copy()
        packet_b = self.copy()
        remain = len(self) - index
        for key in self.gw_data.keys():
            for i in range(index):
                packet_a.gw_data[key] = np.delete(packet_a.gw_data[key], -1)

            for i in range(remain):
                packet_b.gw_data[key] = np.delete(packet_b.gw_data[key], 0)

        return packet_a, packet_b

    def copy(self):
        return copy.deepcopy(self)

    def sort(self):
        """
        sort gw_data lists according to gw_time
        """
        isort = np.argsort(self.gw_data['gw_clock'])
        for key in self.gw_data.keys():
            self.gw_data[key] = self.gw_data[key][isort]

    def get_average_rssi(self):
        return np.average(self.gw_data['rssi'])

    def get_packet_size(self):
        return int(self.packet_data['packet_length'])

    def is_short_packet(self):
        return len(self.get_packet_string(process_packet=False)) < self.get_packet_size()

    def get_packet_content(self, raw_packet=None, get_gw_data=False, sprinkler_num=0):
        if raw_packet is None:
            parsed_packet = self
        else:
            raw_packet = valid_packet_start(raw_packet) or raw_packet
            parsed_packet = Packet(raw_packet=raw_packet)

        if get_gw_data:
            return parsed_packet.packet_data['raw_packet'] + parsed_packet.gw_data['gw_packet'].take(sprinkler_num)

        return parsed_packet.packet_data['raw_packet']

    def get_packet_string(self, i=0, gw_data=True, process_packet=True, full_packet=False):
        """
        gets process_packet string
        """
        if not self.is_valid_packet:
            return None

        process_packet_string = ['', '']
        if process_packet:
            if full_packet:
                raise Exception('Could not get packet string since both full and process packet flags are true')
            process_packet_string = ['process_packet("', '")']
        if full_packet:
            process_packet_string = ['full_packet("', '")']

        if gw_data:
            gw_data = self.gw_data['gw_packet'].take(i)
        else:
            gw_data = ''
        return '{raw_packet}{gw_data}'.format(raw_packet=self.packet_data['raw_packet'],
                                              gw_data=gw_data).join(process_packet_string)

    def get_adva(self):
        return self.packet_data['adv_address']

    def get_flow(self):
        flow_version = hex(int(self.packet_data['flow_ver'], 16))
        return flow_version

    def get_rssi(self):
        return str(self.gw_data['rssi'])

    def get_packet_data_names(self):
        """
        extract all keys name which can potentially be a column in packet df
        :return:
        :rtype: list
        """
        dict_temp = self.__dict__
        keys = list(dict_temp.keys())
        for k in dict_temp.keys():
            if isinstance(dict_temp[k], dict):
                keys = keys + list(dict_temp[k].keys())
        return keys

    def is_same_sprinkler(self, packet):
        raw_packet = packet.packet_data['raw_packet']
        if raw_packet == self.packet_data['raw_packet']:
            return True
        return False

    def append_to_sprinkler(self, packet, output_log=True, do_sort=False):
        status = True
        if self.is_same_sprinkler(packet):
            for i in range(len(packet)):
                try:
                    # stat param should be unique (time + rssi for same packet)- we make sure no duplications are added.
                    if np.take(packet.gw_data['gw_clock'], i) not in self.gw_data['gw_clock'] or \
                            np.take(packet.gw_data['time_from_start'], i) not in self.gw_data['time_from_start']:
                        for key in gw_result_dict.keys():
                            self.gw_data[key] = np.append(self.gw_data[key], np.take(packet.gw_data[key], i))
                        for key in self.custom_data.keys():
                            if key in packet.custom_data.keys():
                                self.custom_data[key] = np.append(self.custom_data[key],
                                                                  np.take(packet.custom_data[key], i))
                            else:
                                self.custom_data[key] = np.append(self.custom_data[key], [None])
                        if do_sort:
                            sort_idx = self.gw_data['time_from_start'].argsort()
                            for k, v in self.gw_data.items():
                                self.gw_data[k] = v[sort_idx]
                            for k, v in self.custom_data.items():
                                self.custom_data[k] = v[sort_idx]
                    else:
                        if output_log:
                            self.printing_and_logging('Tried to add duplicated packet to sprinkler {}'.format(packet))
                except Exception as e:
                    self.printing_and_logging('Failed to add packet {} to sprinkler, exception: {}'
                                              .format(packet, str(e)))
        else:
            self.printing_and_logging('Not from the same sprinkler')
            status = False
        # self.sort()
        return status

    def as_dict(self, sprinkler_index=None):  # None not tested
        packet_data = self.packet_data.copy()
        sprinkler_gw_data = self.gw_data.copy()
        custom_data = self.custom_data.copy()
        if sprinkler_index is not None:
            if sprinkler_index > self.gw_data['stat_param'].size:
                return None
            for gw_attr in gw_result_dict.keys():
                sprinkler_gw_data[gw_attr] = np.take(self.gw_data[gw_attr], sprinkler_index)
            for custom_attr in self.custom_data.keys():
                custom_data[custom_attr] = np.take(self.custom_data[custom_attr], sprinkler_index)
            dict_len = 1
        else:
            dict_len = len(self)
            for k, v in packet_data.items():
                if k == 'raw_packet' and 'adi_per_packet' in custom_data.keys():
                    adi_per_packet_list = custom_data['adi_per_packet']
                    if not isinstance(custom_data['adi_per_packet'], list) and \
                            not isinstance(custom_data['adi_per_packet'], np.ndarray):
                        adi_per_packet_list = [adi_per_packet_list] * dict_len
                    original_raw_packets = [construct_packet_from_fields(
                        packet_data=self.packet_data,
                        fields_to_convert={'adi': adi_per_packet}) for adi_per_packet in adi_per_packet_list]
                    packet_data[k] = original_raw_packets
                    continue
                packet_data[k] = [v] * dict_len

        data = {**packet_data, **sprinkler_gw_data, **custom_data}
        data['gw_process'] = [self.gw_process] * dict_len  # bool should be a list for pd.DataFrame.from_dict(data)
        data['is_valid_packet'] = [self.is_valid_packet] * dict_len  # bool should be a list
        data['inlay_type'] = [self.inlay_type] * dict_len
        return data

    def as_dataframe(self, sprinkler_index=None):
        data = self.as_dict(sprinkler_index=sprinkler_index)
        packet_df = pd.DataFrame.from_dict(data)

        return packet_df

    def get_per(self, expected_sprinkler_count=None):
        """
        Calculates the packet per at the sprinkler
        @param expected_sprinkler_count - The number of packet per sprinkler
        @return packet per at percentage
        """
        if expected_sprinkler_count is None:
            if self.packet_data['test_mode']:
                expected_sprinkler_count = 6
            else:
                expected_sprinkler_count = get_flow_param(flow_in=self.packet_data['flow_ver'],
                                                          param_name='sprinkler_num')
        return 1 - (len(self) / expected_sprinkler_count)

    def get_tbp(self):
        """
        calculates the rate of packets from the same sprinkler
        :return: min_times_found - in msec
        :rtype: int
        """

        def triad_ratio_logic(diff_time_1, diff_time_2, ratio=1.0, error=10):
            """ estimate the time between successive packet according to only 3 packets out of 6 """
            if abs(diff_time_1 - ratio * diff_time_2) <= diff_time_2 / error:
                return True
            elif abs(diff_time_1 - (1 / ratio) * diff_time_2) <= diff_time_1 / error:
                return True
            else:
                return False

        def estimate_diff_packet_time(times_list, pc_time_list):
            if times_list.size < sprinkler_num/2 or times_list.size == 1:
                return None

            # sorting based on the pc time:
            sort_idx = pc_time_list.argsort()
            pc_time_list_sorted = pc_time_list.copy()[sort_idx]
            times_list_sorted = times_list.copy()[sort_idx]

            # fix sorting if several packets have the same timestamp
            fix_sort_idx = sort_idx
            # fix_sort_idx = []
            # ind = 0
            # pc_time_unique, pc_time_counts = np.unique(pc_time_list_sorted, return_counts=True)
            # for unique, counts in zip(pc_time_unique, pc_time_counts):
            #     if counts > 1:  # several packets
            #         ind_with_same_time = [i for i, t in enumerate(pc_time_list_sorted) if t == unique]
            #         fix_sort_idx += list(times_list_sorted[ind_with_same_time].argsort() + ind_with_same_time[0])
            #     else:
            #         fix_sort_idx.append(sort_idx[ind])
            #     ind += counts

            times_list_sorted = times_list_sorted[fix_sort_idx]
            dt = []
            for i, t_hw in enumerate(zip(times_list_sorted[1::], times_list_sorted[:-1])):
                dt_tmp = t_hw[0] - t_hw[1]
                dt_pc = round((pc_time_list_sorted[i + 1] - pc_time_list_sorted[i]) * 1000)
                if dt_pc >= MAX_GW_CLOCK:
                    # HW timing was zeroing during the same packets sprinkler
                    dt_tmp = dt_pc
                elif dt_tmp < 0:
                    dt_tmp += MAX_GW_CLOCK

                dt.append(dt_tmp)

            return dt

        def check_if_nan(lst):
            is_nan = pd.isnull(lst)
            if isinstance(is_nan, bool):
                return is_nan
            if isinstance(is_nan.tolist(), bool):
                return is_nan.tolist()
            if isinstance(is_nan, np.ndarray):
                return any(is_nan)
            return False

        if check_if_nan(self.gw_data['time_from_start']) or check_if_nan(self.gw_data['gw_clock']):
            return None

        if self.packet_data['test_mode']:
            sprinkler_num = 6
        else:
            sprinkler_num = get_flow_param(flow_in=self.packet_data['flow_ver'], param_name='sprinkler_num')
        estimate_diff_time = estimate_diff_packet_time(self.gw_data['gw_clock'], self.gw_data['time_from_start'])
        if estimate_diff_time is None:
            return None
        elif len(self) == sprinkler_num/2 and sprinkler_num == 6:
            if triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=1):
                return None
            else:
                for ratio in [2, 3, 4]:
                    if triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=ratio):
                        estimate_diff_time = [min(estimate_diff_time[0], estimate_diff_time[1]),
                                              max(estimate_diff_time[0], estimate_diff_time[1]) / ratio]
                        break
                if triad_ratio_logic(estimate_diff_time[0], estimate_diff_time[1], ratio=1.5):
                    estimate_diff_time = [min(estimate_diff_time[0], estimate_diff_time[1]) / 2,
                                          max(estimate_diff_time[0], estimate_diff_time[1]) / 3]

        return int(min(estimate_diff_time))

    def extract_packet_data_by_name(self, key):
        """
        extract data from all packet attribute according to the key name
        :param key: the name of the data (the key of the relevant dictionary)
        :type key: str
        :return: list of the data
        :rtype: list
        """

        if key in self.packet_data:
            data_attr = self.packet_data[key]
        elif hasattr(self, 'decoded_data') and key in self.decoded_data:
            data_attr = self.decoded_data[key]
        elif key in self.gw_data:
            data_attr = self.gw_data[key]
        elif key in self.custom_data:
            data_attr = self.custom_data[key]
        else:
            self.printing_and_logging('key:{} does not exist in packet structure')
            return None
        if isinstance(data_attr, np.ndarray):
            data_attr = data_attr.tolist()
        if not isinstance(data_attr, list):
            data_attr = [data_attr]
        return data_attr

    def filter_by_sprinkler_id(self, sprinkler_ids):
        """
        keep only specific sprinklers in a packet according to the sprinkler id
        :param sprinkler_ids:
        :type sprinkler_ids: list
        :return: filtered packet
        :rtype: Packet or DecryptedPacket
        """

        def filter_per_attr(data_attr):
            for key in data_attr.keys():
                if isinstance(data_attr[key], list) or \
                        (isinstance(data_attr[key], np.ndarray) and data_attr[key].size > 1):
                    filtered_data = [data_attr[key][i] for i in sprinkler_ids]
                    if isinstance(data_attr[key], np.ndarray):
                        data_attr[key] = np.array(filtered_data)
                    else:
                        data_attr[key] = filtered_data
            return data_attr

        filtered_packet = self.copy()
        filtered_packet.gw_data = filter_per_attr(filtered_packet.gw_data)
        filtered_packet.custom_data = filter_per_attr(filtered_packet.custom_data)

        return filtered_packet

    def add_custom_data(self, custom_data):
        for key in custom_data.keys():
            if isinstance(custom_data[key], list):
                if len(custom_data[key]) == self.__len__():
                    self.custom_data[key] = custom_data[key]
                else:
                    self.printing_and_logging('add_custom_data failed - '
                                              'the custom data is a list of a different length than'
                                              ' the number of packets')
            else:
                self.custom_data[key] = self.__len__() * [custom_data[key]]


if __name__ == '__main__':

    # test mode packets with different adi, same sprinkler:
    testing_82_tm_adi0 = Packet(
        'full_packet("4731090906D03E34CD8200001EFF000505000060750774415ACDFFFBD069756C550DB1FD75E7B127312C21F3B10A03B1CB12AF2D45AC")')
    testing_82_tm_adi1 = Packet(
        'full_packet("4731090906D03E34CD8200011EFF000505000060750774415ACDFFFBD069756C550DB1FD75E7B127312C21F3B10A03B1CB12AF2D45CE")')
    testing_82_tm_adi2 = Packet(
        'full_packet("4731090906D03E34CD8200031EFF000505000060750774415ACDFFFBD069756C550DB1FD75E7B127312C21F3B10A03B1CB12AF2D4615")')
    testing_82_tm_adi0.append_to_sprinkler(testing_82_tm_adi1)
    testing_82_tm_adi0.append_to_sprinkler(testing_82_tm_adi2)
    print(f'packet len: {len(testing_82_tm_adi0)} with the following adi: '
          f'{testing_82_tm_adi0.custom_data["adi_per_packet"]}')

    packet_1 = {'packet': '03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C404200486436AE8F',
                'is_valid_tag_packet': True, 'adv_address': '03B28DCD9920', 'group_id': 'FE0000', 'rssi': 54,
                'stat_param': 30687,
                'time_from_start': 1.528374, 'counter_tag': 1}
    packet_2 = '03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C404200486436AE82'

    p1 = Packet(packet_1)

    print(p1.get_packet_content('03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C404200486436AE82', get_gw_data=True),)

    p2 = Packet(packet_2)

    p3 = Packet('0437E62F33341E16AFFD02000058AD2AE76DE1D314B003083386A2A1AD139E42C10A0448F93472DC')
    print(p3.get_payload())

    packet_no_gw = Packet('03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C4042004864')
    packet_corrupted_length = Packet('03B28DCD99201EFF0005FE0000E0210FFF93B635EBFF1DB118C6D782DC2ED98C4042004864x')

    print(p1.get_packet_string(0))
    print(p1.get_average_rssi())

    print(p1 == p2)
    print(p1.append_to_sprinkler(p2))

    p6_1 = Packet(raw_packet='473109090673E063E00800002616AFFDFE0080FB7140E0F9B527E8E4751A0A18A99E360281421735AF8A000281415D687CC27F529230',
                  time_from_start=81.092617)
    p6_2 = Packet(raw_packet='473109090673E063E00800002616AFFDFE0080FB7140E0F9B527E8E4751A0A18A99E360281421735AF8A000281415D687CC27F52DEA2',
                  time_from_start=228.187184)
    p6 = p6_1.append_to_sprinkler(p6_2)
    tbp = p6_1.get_tbp()

    p6_3 = Packet(raw_packet='473109090695564AAB0800002616AFFDFE00C05AA9955A56FF3B552A24B28D275BDC8647CFDA0750ABA8541FF99340948F41D450AFCC',
                  time_from_start=38.627151)
    p6_4 = Packet(
        raw_packet='473109090695564AAB0800002616AFFDFE00C05AA9955A56FF3B552A24B28D275BDC8647CFDA0750ABA8541FF99340948F41D451BB8E',
        time_from_start=105.55923)
    p6a = p6_3.append_to_sprinkler(p6_4)
    tbp2 = p6_3.get_tbp()

    p6_5 = Packet(
        raw_packet='473109090695564AAB0800002616AFFDFE00C05AA9955A56FF3B552A24B28D275BDC8647CFDA0750ABA8541FF99340948F41D450FFFC',
        time_from_start=0.0)
    p6_6 = Packet(
        raw_packet='473109090695564AAB0800002616AFFDFE00C05AA9955A56FF3B552A24B28D275BDC8647CFDA0750ABA8541FF99340948F41D4510010',
        time_from_start=10.55923)
    p6b = p6_5.append_to_sprinkler(p6_6)
    tbp3 = p6_5.get_tbp()

    p6_7 = Packet(
        raw_packet='06455AD22E061E16AFFDFE0040FFBD941223357780F3FBC45BA7E13D7E6AFB4F89D61463EA3EFF9A',
        time_from_start=55796.199652)
    p6_8 = Packet(
        raw_packet='06455AD22E061E16AFFDFE0040FFBD941223357780F3FBC45BA7E13D7E6AFB4F89D61463EA3E00DA',
        time_from_start=55796.199652)
    p67 = p6_7.append_to_sprinkler(p6_8)
    tbp4 = p6_7.get_tbp()

    print(len(p1))
    print(p1.get_average_rssi())

    p1_dict = p1.as_dict()
    p1_df = pd.DataFrame(data=p1_dict)
    print(p1_df)

    print('end')



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

# check out the confluence: https://wiliot.atlassian.net/wiki/spaces/SW/pages/3228827713/GW+Tester+Board+GPIO

import time

from wiliot_core import set_logger
from wiliot_core import WiliotGateway, ActionType, DataType
from wiliot_core import PacketList


MASTER_NAME = '3D2C07C18B3BBD54'  # to get gw name you can send !get_name
WORKER_NAME = '6CE32EC8D34804A3'
CONNECTED_PIN = 'P004'  # both GWs should be connect by one of the optional GPIO (check out the above link ) + GND
TEST_TIME = 5  # seconds

if __name__ == '__main__':
    my_logger_path, my_logger = set_logger(app_name='DualGwExample',
                                           dir_name='dual_gw_example',
                                           file_name='dual_gw_log')
    # init:
    gw_master = WiliotGateway(device_name=MASTER_NAME, is_multi_processes=True, logger_name=my_logger.name,
                              mp_reset_time_upon_gw_start=True)
    if not gw_master.is_connected():
        my_logger.error('error during connection to gw master')
        exit(-1)
    gw_master.reset_gw()
    if not gw_master.is_gw_alive():
        my_logger.error('error after reset gw master')
        exit(-1)

    gw_worker = WiliotGateway(device_name=WORKER_NAME, is_multi_processes=True, logger_name=my_logger.name)
    if not gw_worker.is_connected():
        my_logger.error('error during connection to gw worker')
        exit(-1)
    gw_worker.reset_gw()
    if not gw_worker.is_gw_alive():
        my_logger.error('error after reset gw worker')
        exit(-1)

    # configure:
    gw_master.config_gw(received_channel=37, energy_pattern_val=18, time_profile_val=[5, 15],
                        with_ack=True, start_gw_app=False)
    gw_rsp = gw_master.write(f"!cmd_gpio CONTROL_OUT {CONNECTED_PIN} 0")  # both gw are working together
    my_logger.info(gw_rsp)

    gw_worker.config_gw(received_channel=37, time_profile_val=[0, 15], with_ack=True, start_gw_app=False)
    gw_rsp = gw_worker.write(f"!cmd_gpio CONTROL_IN {CONNECTED_PIN} 1")
    my_logger.info(gw_rsp)

    # start collecting packets
    gw_master.config_gw(start_gw_app=True)
    t_i = time.time()
    all_packets_master = PacketList()
    all_packets_worker = PacketList()
    all_worker_rsp = []
    while time.time() - t_i < TEST_TIME:
        # collecting from master
        if gw_master.is_data_available():
            packets_master = gw_master.get_packets(action_type=ActionType.ALL_SAMPLE, data_type=DataType.PACKET_LIST)
            all_packets_master.__add__(packets_master)

        # collecting from slave
        if gw_worker.is_data_available():
            packets_worker = gw_worker.get_packets(action_type=ActionType.ALL_SAMPLE, data_type=DataType.PACKET_LIST)
            all_packets_worker.__add__(packets_worker)

        # for time sync
        all_worker_rsp += gw_worker.get_gw_responses()

    # stop both GWs:
    gw_master.stop_gw_app()
    gw_master.exit_gw_api()

    # merge results
    try:
        time_to_sync = gw_worker.get_time_of_controlled_start_app(all_worker_rsp)
        all_packets_master.sync_packet_list(new_packet_list=all_packets_worker,
                                            new_source_name='ch4', my_source_name='ch37',
                                            time_sync_new=time_to_sync, do_sort=True)
    except Exception as e:
        my_logger.warning(f'could not merge results due to {e}')

    my_logger.info(f'merge packet list: {len(all_packets_master)}')
    df = all_packets_master.get_df(add_sprinkler_info=True)
    df.to_csv('dual_test.csv')
    my_logger.info(df)
    my_logger.info('done')



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import numpy as np
from enum import Enum


class FieldsLength(Enum):
    HEADER = 4
    EXTENDED_HEADER = 4
    ADVA = 12
    ADI = 4
    EN = 2
    TYPE = 2
    DATA_UID = 4
    GROUP_ID = 6
    NONCE = 8
    ENC_UID = 12
    MIC = 12
    ENC_PAYLOAD = 16


class BridgeParams(Enum):
    PACKET_DATA_INFO = 4  # 1E16
    SERVICE_UUID_LENGTH = 4  # AFFD
    BRG_SERVICE_UUID = 'C6FC'


class GWPacketLength(Enum):
    STAT_PARAM_LENGTH = 4
    RSSI_LENGTH = 2
    GW_DATA_LENGTH = STAT_PARAM_LENGTH + RSSI_LENGTH


BITS_IN_CHAR = 4

PACKET_DECRYPT_MASK = 0xc
GROUP_ID_MASK = 0xFFFF00
RAW_GROUP_ID_MASK = 0xFFFFC0

BLE5_SHIFT = FieldsLength.HEADER.value + FieldsLength.EXTENDED_HEADER.value + FieldsLength.ADI.value
ADI_LOCATION = FieldsLength.HEADER.value + FieldsLength.EXTENDED_HEADER.value + FieldsLength.ADVA.value
FLOW_VER_SIZE = 4

PAYLOAD_LENGTH = 16
BLE_PAYLOAD_LENGTH = 62
BLE_DOUBLE_PAYLOAD_LENGTH = 78
WILIOT_EN_TYPE = ['1E16', '2616']
WILIOT_DATA_UID = ['C6FC', 'AFFD']

packet_structure = [{'name': 'header', 'len': FieldsLength.HEADER.value},
                    {'name': 'extended_header', 'len': FieldsLength.EXTENDED_HEADER.value},
                    {'name': 'adv_address', 'len': FieldsLength.ADVA.value},
                    {'name': 'adi', 'len': FieldsLength.ADI.value},
                    {'name': 'en', 'len': FieldsLength.EN.value},
                    {'name': 'type', 'len': FieldsLength.TYPE.value},
                    {'name': 'data_uid', 'len': FieldsLength.DATA_UID.value},
                    {'name': 'raw_group_id', 'len': FieldsLength.GROUP_ID.value},
                    {'name': 'nonce', 'len': FieldsLength.NONCE.value},
                    {'name': 'enc_uid', 'len': FieldsLength.ENC_UID.value},
                    {'name': 'mic', 'len': FieldsLength.MIC.value},
                    {'name': 'enc_payload', 'len': FieldsLength.ENC_PAYLOAD.value}
                    ]
gw_bit_structure = [{'name': 'rssi', 'bit_len': GWPacketLength.RSSI_LENGTH.value * BITS_IN_CHAR},
                    {'name': 'crc_valid', 'bit_len': 1},
                    {'name': 'gw_clock', 'bit_len': GWPacketLength.STAT_PARAM_LENGTH.value * BITS_IN_CHAR - 1},
                    ]
BLE5_ONLY_FIELDS = ['header', 'extended_header', 'adi']

packet_data_dict = {'raw_packet': '',
                    'adv_address': (0, 12),
                    'decrypted_packet_type': (24, 1),
                    'group_id': (20, 6),
                    'raw_group_id': '',
                    'test_mode': 0,
                    'header': '',
                    'extended_header': '',
                    'adi': '',
                    'en': (12, 2),
                    'type': (14, 2),
                    'data_uid': (16, 4),
                    'nonce': (26, 8),
                    'enc_uid': (34, 12),
                    'mic': (46, 12),
                    'enc_payload': (58, 16),
                    'packet_length': None}

gw_result_dict = {'gw_packet': 'gw_packet',
                  'rssi': 'rssi',
                  'stat_param': 'stat_param',
                  'crc_valid': (0, 1),  # bits in stat param
                  'gw_clock': (1, 16),  # bits in stat param
                  'time_from_start': 'time_from_start',
                  'counter_tag': 'counter_tag',
                  'is_valid_tag_packet': 'is_valid_tag_packet',
                  'is_packet_from_bridge': 'is_packet_from_bridge'}

packet_length_types = {
    '4225': {'name': 'LEGACY', 'packet_tag_length': 78, 'bytes_shift': 0, 'length_modifier': None},
    '4729': {'name': 'BLE5-EXT', 'packet_tag_length': 86, 'bytes_shift': BLE5_SHIFT, 'length_modifier': None},
    '4731': {'name': 'BLE5-DBL-EXT', 'packet_tag_length': 102, 'bytes_shift': BLE5_SHIFT,
             'length_modifier': {'enc_payload': (PAYLOAD_LENGTH * 2)}}
}


def parse_packet(packet_data_input, is_full_packet, ignore_crc):
    pixie_result = {}
    gw_result = {}

    # check packet type:
    for full_key in packet_length_types.keys():
        if packet_data_input.startswith(full_key):
            is_full_packet = True

    if is_full_packet is None:
        packet_data_input, expected_length = check_packet_length_with_no_indication(packet_data_input)
        if expected_length is None:
            raise Exception(f'invalid packet length for packet {packet_data_input}, '
                            f'these are the valid tag packet length: {packet_length_types}')
    elif is_full_packet:
        expected_length = packet_length_types[packet_data_input[:FieldsLength.HEADER.value]]
    else:
        packet_data_input = '4225' + packet_data_input
        expected_length = packet_length_types[packet_data_input[:FieldsLength.HEADER.value]]

    # check if packet from bridge
    from_bridge = is_packet_from_bridge(packet_data_input)

    # check if packet length is valid
    received_len = len(packet_data_input)

    if received_len == expected_length['packet_tag_length'] + GWPacketLength.GW_DATA_LENGTH.value:
        pass  # valid length
    elif received_len == expected_length['packet_tag_length']:
        packet_data_input += '0' * GWPacketLength.GW_DATA_LENGTH.value
    else:
        gw_result['is_valid_tag_packet'] = np.array(False)
        raise Exception(f'invalid packet length for packet {packet_data_input}, '
                        f'expected tag packet length: {expected_length["packet_tag_length"]}')

    try:
        # Parse pixie data
        pixie_result.update(parse_pixie(packet_data_input, expected_length, from_bridge))

        # Parse GW
        gw_result.update(parse_gw(packet_data_input, from_bridge, ignore_crc))
        gw_result['is_valid_tag_packet'] = np.array(True)

    except Exception as e:
        print('Packet string cannot be parsed due to {}'.format(e))
        gw_result['is_valid_tag_packet'] = np.array(False)

    return pixie_result, gw_result


def parse_gw(packet_data, is_from_bridge, ignore_crc):
    result = {}
    gw_data = packet_data[-GWPacketLength.GW_DATA_LENGTH.value:]
    try:
        result['gw_packet'] = np.array(gw_data, dtype='<U6')
        result['rssi'] = np.array(int(gw_data[:GWPacketLength.RSSI_LENGTH.value], 16))
        result['stat_param'] = np.array(int(gw_data[GWPacketLength.RSSI_LENGTH.value:
                                                    GWPacketLength.RSSI_LENGTH.value +
                                                    GWPacketLength.STAT_PARAM_LENGTH.value], 16))
        stat_param_bits = format(result['stat_param'].item(), '04b').zfill(
            GWPacketLength.STAT_PARAM_LENGTH.value * BITS_IN_CHAR)
        if ignore_crc:
            result['crc_valid'] = np.array(1)
        else:
            result['crc_valid'] = np.array(int(stat_param_bits[
                                           gw_result_dict['crc_valid'][0]:gw_result_dict['crc_valid'][1]]))
        result['gw_clock'] = np.array(int(stat_param_bits[
                                          gw_result_dict['gw_clock'][0]:gw_result_dict['gw_clock'][1]], 2))
        result['time_from_start'] = np.array(float('nan'))
        result['counter_tag'] = np.array(float('nan'))
        result['is_packet_from_bridge'] = np.array(is_from_bridge)

        return result

    except Exception as e:
        print('Issue parsing GW data: {}'.format(e))
        return


def parse_pixie(packet_data, packet_len_dict, is_from_bridge):
    try:
        result = {'raw_packet': packet_data}
        valid_length = packet_len_dict['packet_tag_length'] + GWPacketLength.GW_DATA_LENGTH.value
        start_index = FieldsLength.HEADER.value if packet_len_dict['name'] == 'LEGACY' else 0

        result['header'] = packet_data[:FieldsLength.HEADER.value]
        result['raw_packet'] = packet_data[start_index:-GWPacketLength.GW_DATA_LENGTH.value]
        result['packet_length'] = valid_length
        result.update(parser(result['raw_packet'], packet_len_dict, is_from_bridge))
        result['decrypted_packet_type'] = (int(result['decrypted_packet_type'], 16) & PACKET_DECRYPT_MASK) >> 2

    except Exception as e:
        raise Exception(f'packet_map: parse_packet: could not parse packet: {packet_data} due to {e}')

    return result


def extract_fields(packet_data, shift=0, length_modifier=None):
    result = {}
    for key, value in packet_data_dict.items():
        if isinstance(value, tuple):
            start, length = value

            if length_modifier and key in length_modifier:
                length = length_modifier[key]

            start_ind = start + shift
            if shift == BLE5_SHIFT and key == 'adv_address':
                start_ind -= FieldsLength.ADI.value

            result[key] = packet_data[start_ind:start_ind + length]
            if key == 'group_id':
                result['raw_group_id'] = \
                    hex(int(result[key], 16) & RAW_GROUP_ID_MASK)[2:].zfill(FieldsLength.GROUP_ID.value).upper()
                result[key] = hex(int(result[key], 16) & GROUP_ID_MASK)[2:].zfill(FieldsLength.GROUP_ID.value).upper()

    return result


def parser(packet_data, len_dict, is_from_bridge):
    result = extract_fields(packet_data, len_dict['bytes_shift'], len_dict['length_modifier'])
    if (is_from_bridge and '**' not in result['adv_address']) or result['adv_address'].startswith('***'):
        result['flow_ver'] = hex(0)
    else:
        result['flow_ver'] = hex(int(result['adv_address'][:2] + result['adv_address'][-2:], 16))
    result['ble_type'] = len_dict['name']
    test_mode = test_mode_check(result)
    result['test_mode'] = test_mode
    if len_dict['name'] != 'LEGACY':
        result['extended_header'] = packet_data[FieldsLength.HEADER.value:
                                                FieldsLength.HEADER.value + FieldsLength.EXTENDED_HEADER.value]
        result['adi'] = packet_data[ADI_LOCATION:ADI_LOCATION + FieldsLength.ADI.value]

    return result


def test_mode_check(pixie_dict):
    flow_version = hex(int(pixie_dict.get('flow_ver', '0x0'), 16))

    if int(flow_version, 16) < 0x42c:
        if 'FFFFFFFF' in pixie_dict.get('adv_address', ''):
            return 1
    elif int(flow_version, 16) < 0x500:
        adv_address = pixie_dict.get('adv_address', '')
        if adv_address.startswith('FFFF') or adv_address.endswith('FFFF'):
            return 1
    else:
        if int(pixie_dict.get('data_uid', '0'), 16) == 5:
            return 1
    return 0


def is_packet_from_bridge(packet_data):
    service_uuid_index = FieldsLength.HEADER.value + FieldsLength.ADVA.value + BridgeParams.PACKET_DATA_INFO.value
    service_uuid = packet_data[service_uuid_index: service_uuid_index + BridgeParams.SERVICE_UUID_LENGTH.value]
    return service_uuid == BridgeParams.BRG_SERVICE_UUID.value


def extract_group_id(raw_group_id):
    """
    Extract group ID from the raw group ID packet by removing the two last bits.
    :param raw_group_id: as it received by the gateway
    :type raw_group_id: str
    :return: the group ID
    :rtype: str
    """
    raw_group_id_list = [x for x in raw_group_id]
    last_half_byte = raw_group_id_list[4]
    byte_without_2_last_bits = int(last_half_byte, 16) & 3
    raw_group_id_list[4] = str(byte_without_2_last_bits)
    return ''.join(raw_group_id_list)


def hex2bin(hex_value, min_digits=0, zfill=True):
    binary_value = format(int(hex_value, 16), 'b')

    if zfill:
        binary_value = binary_value.zfill(24)

    if len(binary_value) < min_digits:
        binary_value = binary_value.zfill(min_digits)

    return binary_value


def check_packet_length_with_no_indication(packet_data_input):
    expected_length = None
    received_len = len(packet_data_input)
    for packet_prefix, length_dict in packet_length_types.items():
        if received_len == length_dict['packet_tag_length']:
            packet_data_input += '0' * GWPacketLength.GW_DATA_LENGTH.value
            expected_length = length_dict

        elif received_len == length_dict['packet_tag_length'] + GWPacketLength.GW_DATA_LENGTH.value:
            expected_length = length_dict

        elif length_dict['name'] == 'LEGACY':
            if received_len == length_dict['packet_tag_length'] - FieldsLength.HEADER.value:
                packet_data_input = packet_prefix + packet_data_input + '0' * GWPacketLength.GW_DATA_LENGTH.value
                expected_length = length_dict
            elif received_len == \
                    length_dict['packet_tag_length'] - FieldsLength.HEADER.value + GWPacketLength.GW_DATA_LENGTH.value:
                packet_data_input = packet_prefix + packet_data_input
                expected_length = length_dict

        if expected_length is not None:
            break

    return packet_data_input, expected_length


def construct_packet_from_fields(packet_data, fields_to_zero=None, fields_to_convert=None, gw_data='', prefix=False,
                                 gw_fields_to_convert=None):
    """
    fields_to_zero : a list of fields names, need to be zero
    fields_to_convert : dict of field name as key and its value based on packet_structure variable
    gw_fields_to_convert : if gw_data is not specified,
                           dict of field name as key and its value based on gw_bit_structure variable
    """
    new_packet = []
    is_legacy_packet = packet_length_types.get(packet_data['header'], {'name': 'LEGACY'})['name'] == 'LEGACY'
    for field in packet_structure:
        if is_legacy_packet and field['name'] in BLE5_ONLY_FIELDS:
            continue  # do not add ble5 fields to legacy packets
        if fields_to_zero and field['name'] in fields_to_zero:
            new_packet.append('0' * field['len'])
        elif fields_to_convert and field['name'] in fields_to_convert:
            new_packet.append(fields_to_convert[field['name']].upper())
        else:
            new_packet.append(packet_data.get(field['name'], ''))

    construct_packet = ''.join(new_packet)
    if gw_data == '' and gw_fields_to_convert is not None:
        gw_bit_data = ''
        if 'crc_valid' not in gw_fields_to_convert.keys():
            gw_fields_to_convert['crc_valid'] = 1
        for field in gw_bit_structure:
            if field['name'] in gw_fields_to_convert.keys():
                bin_field = bin(int(gw_fields_to_convert[field['name']]))[2:]
            else:
                bin_field = '0'
            bin_field = bin_field.zfill(field['bit_len'])
            bin_field = bin_field[-field['bit_len']:]
            gw_bit_data += str(bin_field)
        gw_data = hex(int(gw_bit_data, 2))[2:].upper()

    construct_packet += gw_data

    if prefix:
        construct_packet = f'process_packet("{construct_packet}")' if is_legacy_packet \
            else f'full_packet("{construct_packet}")'
    return construct_packet


def convert_cloud_to_packet(cloud_packet, original_packet=None):
    if original_packet is None:
        adva = ''.join(['*'] * FieldsLength.ADVA.value)
        en = ''.join(['*'] * FieldsLength.EN.value)
        b_type = ''.join(['*'] * FieldsLength.TYPE.value)
        e_header = ''.join(['*']) * FieldsLength.EXTENDED_HEADER.value
        adi = ''.join(['*']) * FieldsLength.ADI.value
        gw_packet = ''
    else:
        adva = original_packet.packet_data['adv_address']
        en = original_packet.packet_data['en']
        b_type = original_packet.packet_data['type']
        e_header = original_packet.packet_data['extended_header'] \
            if 'extended_header' in original_packet.packet_data.keys() else \
            ''.join(['*']) * FieldsLength.EXTENDED_HEADER.value
        adi = original_packet.packet_data['adi'] \
            if 'adi' in original_packet.packet_data.keys() else \
            ''.join(['*']) * FieldsLength.ADI.value
        gw_packet = np.take(original_packet.gw_data['gw_packet'], 0)

    if any(cloud_packet.startswith(x) for x in WILIOT_EN_TYPE) and len(cloud_packet) == BLE_PAYLOAD_LENGTH:
        return adva + cloud_packet + gw_packet
    if any(cloud_packet.startswith(x) for x in WILIOT_DATA_UID) and \
            len(cloud_packet) == BLE_PAYLOAD_LENGTH - FieldsLength.EN.value - FieldsLength.TYPE.value:
        return adva + en + b_type + cloud_packet + gw_packet

    if any(cloud_packet.startswith(x) for x in WILIOT_EN_TYPE) and \
            len(cloud_packet) == BLE_DOUBLE_PAYLOAD_LENGTH:
        header = [k for k, v in packet_length_types.items() if v['name'] == 'BLE5-DBL-EXT']
        return header[0] + adva + e_header + adi + cloud_packet + gw_packet
    if any(cloud_packet.startswith(x) for x in WILIOT_DATA_UID) and \
            len(cloud_packet) == BLE_DOUBLE_PAYLOAD_LENGTH - FieldsLength.EN.value - FieldsLength.TYPE.value:
        header = [k for k, v in packet_length_types.items() if v['name'] == 'BLE5-DBL-EXT']
        return header[0] + adva + e_header + adi + en + b_type + cloud_packet + gw_packet

    return cloud_packet


def reverse_data(data):
    data_reversed = ''
    for i in range(len(data), 0, -2):
        data_reversed += data[i - 2:i]
    return data_reversed



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

from wiliot_core.packet_data.config_files.packet_map import *

N_BITS_PER_CHAR = 4


class ConvertCloudPacket(object):
    def __init__(self, enriched_packet_row, original_packet=None, first_timestamp=0):
        """
        :param enriched_packet_row is one row from the enrich packet table from DB
        :type enriched_packet_row pd.Series
        """
        self.row = enriched_packet_row
        self.first_time = first_timestamp
        if 'rawPacket' in self.row.keys() and ('packetVersion' in self.row.keys() or 'flowVersion' in self.row.keys()):
            pass
        else:
            raise Exception('cannot convert cloud packet without the following fields: '
                            '[rawPacket and packetVersion] or [rawPacket and flowVersion]')

        flow_ver = self.get_flow_ver()
        self.gw_packet = self.get_gw_data()
        base_fields = self.get_fields_before_payload(original_packet=original_packet, flow_ver=flow_ver)
        packet_out = self.combined_packet_parts(base_fields)
        packet_fixed = self.get_packet_data(packet_out)
        self.base_fields = base_fields
        self.reconstruct_packet = packet_fixed

    def get_reconstruct_packet(self):
        return self.reconstruct_packet

    def get_flow_ver(self):
        flow_ver_hex = '****'
        if 'flowVersion' in self.row:
            flow_ver_hex = str(self.row['flowVersion'])
            flow_ver_hex = flow_ver_hex.upper().replace('.', '').replace('X', '').zfill(FLOW_VER_SIZE)
            if len(flow_ver_hex) != FLOW_VER_SIZE:
                raise Exception(f'ConvertCloudPacket: get_flow_ver: invlid flow version: {self.row["flowVersion"]}')
        return flow_ver_hex

    def get_fields_before_payload(self, original_packet, flow_ver):
        if original_packet is None:
            adva = flow_ver[:int(FLOW_VER_SIZE/2)] + \
                   ''.join(['*'] * (FieldsLength.ADVA.value - FLOW_VER_SIZE)) + \
                   flow_ver[-int(FLOW_VER_SIZE/2):]
            en = ''.join(['*'] * FieldsLength.EN.value)
            b_type = ''.join(['*'] * FieldsLength.TYPE.value)
            e_header = ''.join(['*']) * FieldsLength.EXTENDED_HEADER.value
            adi = ''.join(['*']) * FieldsLength.ADI.value
            gw_packet = self.gw_packet
        else:
            adva = original_packet.packet_data['adv_address']
            en = original_packet.packet_data['en']
            b_type = original_packet.packet_data['type']
            e_header = original_packet.packet_data['extended_header'] \
                if 'extended_header' in original_packet.packet_data.keys() else \
                ''.join(['*']) * FieldsLength.EXTENDED_HEADER.value
            adi = original_packet.packet_data['adi'] \
                if 'adi' in original_packet.packet_data.keys() else \
                ''.join(['*']) * FieldsLength.ADI.value
            gw_packet = np.take(original_packet.gw_data['gw_packet'], 0)

        return {'adva': adva, 'en': en, 'b_type': b_type, 'e_header': e_header, 'adi': adi, 'gw_packet': gw_packet}

    def get_gw_data(self):
        gw_fields = {}
        gw_data = ''
        if 'rssi' in self.row:
            gw_fields['rssi'] = abs(int(self.row['rssi']))
        if 'timestamp' in self.row:
            gw_fields['gw_clock'] = int(self.row['timestamp']) - self.first_time

        if gw_fields:
            gw_bit_data = ''
            gw_fields['crc_valid'] = 1
            for field in gw_bit_structure:
                if field['name'] in gw_fields.keys():
                    bin_field = bin(int(gw_fields[field['name']]))[2:]
                else:
                    bin_field = '0'
                bin_field = bin_field.zfill(field['bit_len'])
                bin_field = bin_field[-field['bit_len']:]
                gw_bit_data += str(bin_field)
            gw_data = hex(int(gw_bit_data, 2))[2:].upper().zfill(
                int(sum(f['bit_len'] for f in gw_bit_structure) / N_BITS_PER_CHAR))
        return gw_data

    def combined_packet_parts(self, all_fields):
        cloud_packet = str(self.row['rawPacket']).upper()

        if any(cloud_packet.startswith(x) for x in WILIOT_EN_TYPE) and len(cloud_packet) == BLE_PAYLOAD_LENGTH:

            return all_fields['adva'] + cloud_packet + all_fields['gw_packet']

        if any(cloud_packet.startswith(x) for x in WILIOT_DATA_UID) and \
                len(cloud_packet) == BLE_PAYLOAD_LENGTH - FieldsLength.EN.value - FieldsLength.TYPE.value:

            return all_fields['adva'] + all_fields['en'] + all_fields['b_type'] + cloud_packet + all_fields['gw_packet']

        if any(cloud_packet.startswith(x) for x in WILIOT_EN_TYPE) and \
                len(cloud_packet) == BLE_DOUBLE_PAYLOAD_LENGTH:
            header = [k for k, v in packet_length_types.items() if v['name'] == 'BLE5-EXT']

            return header[0] + all_fields['adva'] + all_fields['e_header'] + all_fields[
                'adi'] + cloud_packet + all_fields['gw_packet']

        if any(cloud_packet.startswith(x) for x in WILIOT_DATA_UID) and \
                len(cloud_packet) == BLE_DOUBLE_PAYLOAD_LENGTH - FieldsLength.EN.value - FieldsLength.TYPE.value:
            header = [k for k, v in packet_length_types.items() if v['name'] == 'BLE5-DBL-EXT']

            return header[0] + all_fields['adva'] + all_fields['e_header'] + all_fields['adi'] + all_fields[
                'en'] + all_fields['b_type'] + cloud_packet + all_fields['gw_packet']

    def get_packet_data(self, packet_out):
        packet_data, gw_data = parse_packet(packet_data_input=packet_out, is_full_packet=None, ignore_crc=False)
        fields_to_zero, fields_to_convert = None, None
        if 'decryptedData' in self.row.keys() and 'tagId' in self.row.keys():
            fields_to_zero = ['mic']
            dec_tag_id = reverse_data(str(self.row['tagId']))
            fields_to_convert = {'enc_uid': dec_tag_id,
                                 'enc_payload': str(self.row['decryptedData']).upper()}
        new_packet = construct_packet_from_fields(packet_data=packet_data,
                                                  fields_to_zero=fields_to_zero,
                                                  fields_to_convert=fields_to_convert,
                                                  gw_data=self.gw_packet)
        return new_packet



from enum import Enum


class CommandDetails(Enum):
    version = {'cmd': 'version', 'cmd_shortcut': 'v', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_version'}
    set_energizing_pattern = {'cmd': 'set_energizing_pattern', 'cmd_shortcut': 'sep', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_set_energy_pattern'}
    set_dyn_energizing_pattern = {'cmd': 'set_dyn_energizing_pattern', 'cmd_shortcut': 'sdep', 'num_of_args': ['SKIP_ARGS_VALIDATION'], 'status_return': True, 'cmd_func': 'cmd_set_dynamic_energy_pattern'}
    set_beacons_pattern = {'cmd': 'set_beacons_pattern', 'cmd_shortcut': 'sbp', 'num_of_args': ['SKIP_ARGS_VALIDATION'], 'status_return': True, 'cmd_func': 'cmd_set_beacons_pattern'}
    output_power = {'cmd': 'output_power', 'cmd_shortcut': 'op', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_2_4_ghz_power_set'}
    time_profile = {'cmd': 'time_profile', 'cmd_shortcut': 'tp', 'num_of_args': ['2'], 'status_return': True, 'cmd_func': 'cmd_time_profile_set'}
    set_pacer_interval = {'cmd': 'set_pacer_interval', 'cmd_shortcut': 'pi', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_wlt_pacer_interval_set'}
    set_sub_1_ghz_energizing_frequency = {'cmd': 'set_sub_1_ghz_energizing_frequency', 'cmd_shortcut': 'sf', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_set_sub_1_ghz_energizing_frequency'}
    set_sub_1_ghz_power = {'cmd': 'set_sub_1_ghz_power', 'cmd_shortcut': 'sop', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_sub_1_ghz_power_set'}
    start_2_4_ghz_energizing = {'cmd': 'start_2_4_ghz_energizing', 'cmd_shortcut': 'NONE', 'num_of_args': ['2'], 'status_return': True, 'cmd_func': 'cmd_2_4_ghz_energizing_start'}
    enable_hw_dual_band = {'cmd': 'enable_hw_dual_band', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': False, 'cmd_func': 'cmd_enable_hw_dual_band'}
    gateway_app = {'cmd': 'gateway_app', 'cmd_shortcut': 's', 'num_of_args': ['0', '4'], 'status_return': True, 'cmd_func': 'cmd_gateway_application_start'}
    store_to_flash = {'cmd': 'store_to_flash', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': True, 'cmd_func': 'cmd_store_to_flash'}
    reset = {'cmd': 'reset', 'cmd_shortcut': 'r', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_reset'}
    move_to_bootloader = {'cmd': 'move_to_bootloader', 'cmd_shortcut': 'mtb', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_move_to_bootloader'}
    print_config = {'cmd': 'print_config', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_print_config'}
    print_config_extended = {'cmd': 'print_config_extended', 'cmd_shortcut': 'pce', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_print_config_extended'}
    get_device_address = {'cmd': 'get_device_address', 'cmd_shortcut': 'pda', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_get_device_address'}
    pl_gw_config = {'cmd': 'pl_gw_config', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_pl_gw_config'}
    full_packet_mode = {'cmd': 'full_packet_mode', 'cmd_shortcut': 'fpm', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_full_packet_mode'}
    enable_crc = {'cmd': 'enable_crc', 'cmd_shortcut': 'ecrc', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_enable_wrong_crc'}
    print_crc = {'cmd': 'print_crc', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_print_crc'}
    reset_crc = {'cmd': 'reset_crc', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': True, 'cmd_func': 'cmd_reset_crc_counters'}
    listen_to_tag_only = {'cmd': 'listen_to_tag_only', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_listen_to_tag_only'}
    set_tester_mode = {'cmd': 'set_tester_mode', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_set_tester_mode'}
    beacons_backoff = {'cmd': 'beacons_backoff', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_beacons_backoff_set'}
    scan_ch = {'cmd': 'scan_ch', 'cmd_shortcut': 'NONE', 'num_of_args': ['1', '2'], 'status_return': True, 'cmd_func': 'cmd_scan_channel_set'}
    start_rx = {'cmd': 'start_rx', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': True, 'cmd_func': 'cmd_rx_start'}
    start_cw = {'cmd': 'start_cw', 'cmd_shortcut': 'NONE', 'num_of_args': ['2'], 'status_return': True, 'cmd_func': 'cmd_cw_transmit_time_start'}
    start_tx_carrier = {'cmd': 'start_tx_carrier', 'cmd_shortcut': 'NONE', 'num_of_args': ['0', '1'], 'status_return': True, 'cmd_func': 'cmd_tx_carrier_start'}
    set_packet_filter_on = {'cmd': 'set_packet_filter_on', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': True, 'cmd_func': 'cmd_set_packet_filter_on'}
    set_packet_filter_off = {'cmd': 'set_packet_filter_off', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': True, 'cmd_func': 'cmd_set_packet_filter_off'}
    bypass_pa = {'cmd': 'bypass_pa', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_bypass_pa'}
    bypass_tx_rx = {'cmd': 'bypass_tx_rx', 'cmd_shortcut': 'NONE', 'num_of_args': ['2'], 'status_return': True, 'cmd_func': 'cmd_bypass_tx_rx'}
    trigger_pl = {'cmd': 'trigger_pl', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_trigger_pl'}
    set_pl_delay = {'cmd': 'set_pl_delay', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_set_pl_delay'}
    set_encryption_keys = {'cmd': 'set_encryption_keys', 'cmd_shortcut': 'NONE', 'num_of_args': ['3'], 'status_return': True, 'cmd_func': 'cmd_set_encryption_keys'}
    decrypt_packet = {'cmd': 'decrypt_packet', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': False, 'cmd_func': 'cmd_decrypt_packet'}
    set_rssi_th = {'cmd': 'set_rssi_th', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_set_rssi_th'}
    enable_hv_gpio = {'cmd': 'enable_hv_gpio', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': True, 'cmd_func': 'cmd_enable_hv_gpio'}
    run_ed_sample = {'cmd': 'run_ed_sample', 'cmd_shortcut': 'NONE', 'num_of_args': ['2'], 'status_return': False, 'cmd_func': 'cmd_run_ed_sample'}
    set_lbt = {'cmd': 'set_lbt', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_set_lbt_mode'}
    cancel = {'cmd': 'cancel', 'cmd_shortcut': 'c', 'num_of_args': ['0'], 'status_return': True, 'cmd_func': 'cmd_cancel'}
    sub1g_sync = {'cmd': 'sub1g_sync', 'cmd_shortcut': 'NONE', 'num_of_args': ['1'], 'status_return': True, 'cmd_func': 'cmd_sub1g_sync'}
    beacons_train = {'cmd': 'beacons_train', 'cmd_shortcut': 'bt', 'num_of_args': ['SKIP_ARGS_VALIDATION'], 'status_return': True, 'cmd_func': 'cmd_set_beacons_train'}
    cmd_gpio = {'cmd': 'cmd_gpio', 'cmd_shortcut': 'NONE', 'num_of_args': ['SKIP_ARGS_VALIDATION'], 'status_return': True, 'cmd_func': 'cmd_gpio'}
    get_name = {'cmd': 'get_name', 'cmd_shortcut': 'NONE', 'num_of_args': ['0'], 'status_return': False, 'cmd_func': 'cmd_get_name'}
    set_scan_radio = {'cmd': 'set_scan_radio', 'cmd_shortcut': 'NONE', 'num_of_args': ['2'], 'status_return': True, 'cmd_func': 'cmd_scan_radio_params'}



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import numpy as np
import pandas as pd


class StatisticsCalc(object):
    def __init__(self, packet_df, sprinkler_df):
        """
        :type packet_df pd.DateFrame
        :type sprinkler_df pd.DataFrame
        """
        self.packet_df = packet_df[packet_df['crc_valid'] == 1]
        self.sprinkler_df = sprinkler_df[sprinkler_df['crc_valid'] == 1]
        self.packet_df.reset_index(inplace=True, drop=True)
        self.sprinkler_df.reset_index(inplace=True, drop=True)
        self.n_packets = self.packet_df.shape[0]
        self.n_cycles = self.sprinkler_df.shape[0]
        self.n_bad_packets = (packet_df['crc_valid'] == 0).sum()

    @staticmethod
    def _get_stat(col_in, nan_val=None, added_val=None, added_n=None):
        """
        :param col_in the array for the statistics
        :type col_in pd.Series or np.ndarray
        :param nan_val value of the nan statistics, the default is float('nan')
        :type None or float or int
        :param added_val list of values to add to the statistics, instead of add it to the col_in to improve performance
        :type added_val list
        :param added_n corresponding to added_val, list of the amount of each value from add_val that is needed to be
                       add to the statistics, instead of add it to the col_in to improve performance
        :type added_n list
        """
        if nan_val is None:
            nan_val = float('nan')

        if all(pd.isnull(col_in)):
            return {'mean': nan_val,
                    'std': nan_val,
                    'min': nan_val,
                    'max': nan_val}
        if added_val is None or added_val == [] or added_n is None or added_n == []:
            return {'mean': np.nanmean(col_in),
                    'std': np.nanstd(col_in),
                    'min': np.nanmin(col_in),
                    'max': np.nanmax(col_in)}
        n = sum(added_n + [np.count_nonzero(~np.isnan(col_in))])
        total_val = sum([a_v * a_n for a_v, a_n in zip(added_n, added_val)])
        return {'mean':  (np.nansum(col_in) + total_val) / n,
                'std': -1.0,
                'min': min([np.nanmin(col_in)] + added_val),
                'max': max([np.nanmax(col_in)] + added_val)}

    def _get_unique_list(self, param, param_name=None):
        if param_name is None:
            param_name = param
        return {param_name: ','.join(self.sprinkler_df[param].unique().astype(str))}

    def get_counting(self):
        return {'num_packets': self.n_packets,
                'num_cycles': self.n_cycles,
                'num_bad_crc_packets': self.n_bad_packets
                }

    def get_sprinkler_counter_stat(self):
        stat = self._get_stat(self.sprinkler_df['num_packets_per_sprinkler'])
        return {f'sprinkler_counter_{k}': v for k, v in stat.items()}

    def get_tbp_stat(self):
        tbp_for_calc = self.sprinkler_df['tbp'].replace(to_replace=-1, value=np.nan, inplace=False)
        stat = self._get_stat(tbp_for_calc, -1)
        stat_out = {f'tbp_{k}': v for k, v in stat.items()}
        stat_out['tbp_num_vals'] = sum(~(pd.isnull(tbp_for_calc)))
        return stat_out

    def get_per_stat(self):
        stat = self._get_stat(self.sprinkler_df['per'])
        stat_out = {f'per_{k}': v for k, v in stat.items() if k in ['mean', 'std']}
        return stat_out

    def get_rssi_stat(self):
        stat = self._get_stat(self.packet_df['rssi'])
        return {f'rssi_{k}': v for k, v in stat.items()}

    def get_time_stat(self):
        stat = self._get_stat(self.packet_df['time_from_start'])
        stat_out = {'ttfp': stat['min'],
                    'end_time': stat['max'],
                    'duration': stat['max'] - stat['min']}
        if stat_out.get('duration', 0) > 0:
            stat_out['rx_rate'] = self.n_packets / stat_out['end_time']
            stat_out['rx_rate_normalized'] = self.n_packets / stat_out['duration']
        return stat_out

    def get_flow_ver(self):
        return self._get_unique_list(param='flow_ver')

    def get_asset_id(self):
        if 'asset_id' in self.sprinkler_df:
            return self._get_unique_list(param='asset_id', param_name='asset_id')
        if 'assetId' in self.sprinkler_df:
            return self._get_unique_list(param='assetId', param_name='asset_id')
        return {}
